# server to receive command from client

import socket
import os
import pathlib
import json
import threading

from monitor.config import Config
from monitor.log import Logging


class ServerCommand(object):
    server_command_ = {}

    @staticmethod
    def get():
        return ServerCommand.server_command_

    @staticmethod
    def add_server_command(name, help=""):
        def decorator(func):
            ServerCommand.server_command_[name] = {
                "command": func,
                "help": help,
            }
            return func

        return decorator


class ServerParams(object):
    should_stop = False


def get_config_string(config):
    config = Config.stringify_config(config)
    config_str = json.dumps(config, indent=4)
    Logging.info(f"Sending config {config_str}")
    return config_str


@ServerCommand.add_server_command("get_config", "Get config")
def get_config_server():
    Logging.info("Sending config")
    return get_config_string(Config.get())


@ServerCommand.add_server_command("set_config", "Set config [key1] [key2] ... [value]")
def set_config_server(*args):
    Logging.info("Setting config")
    if len(args) < 2:
        raise Exception("Not enough arguments")
    current = Config.get()
    current_header = current
    keys = args[:-1]
    value = args[-1]
    for key in keys[:-1]:
        if key not in current or not isinstance(current[key], dict):
            raise Exception(f"Key {key} not found")
        current = current[key]
    current[keys[-1]] = value
    current_header = Config.initialize_config(current_header, force=True)
    config_str = json.dumps(current_header, indent=4)
    Logging.info(f"Sending config {config_str}")
    Config.set(Config.update_config(Config.get(), current_header))
    return config_str


@ServerCommand.add_server_command("stop", "Stop server")
def stop_server():
    Logging.info("Stopping")
    ServerParams.should_stop = True
    exit(0)


def should_stop():
    return ServerParams.should_stop


@ServerCommand.add_server_command("restart", "Restart server")
def restart_server():
    Logging.info("Restarting")
    import sys

    this_command = " ".join(["python3"] + sys.argv)
    this_pid = os.getpid()

    os.system(this_command + " &")
    os.system(f"kill {this_pid}")


@ServerCommand.add_server_command("status", "Get status")
def status_server():
    Logging.info("Getting status")
    from monitor.capture import Capture
    from monitor.video import Video
    from monitor.policy import PolicyParams

    return f"captured_frames: {Capture.captured_frames()}/{PolicyParams.frames_count} generated_videos: {len(Video.generated_videos())}"


def start_server():
    # create a socket object
    serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # get local machine name
    host = socket.gethostname()
    port = Config.get()["server"]["port"]

    # force to release the port
    serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # bind to the port
    serversocket.bind((host, port))

    # queue up to 5 requests
    serversocket.listen(5)

    def server_loop():
        Logging.info("Server started")
        server_command = ServerCommand.get()
        while True:
            # establish a connection
            try:
                serversocket.settimeout(60)
                clientsocket, addr = serversocket.accept()
            except socket.timeout:
                continue

            Logging.info(f"Got a connection from {addr}")
            msg = clientsocket.recv(1024).decode("utf-8").split()
            Logging.info(f"Received {msg}")
            argument = msg[1:]
            msg = msg[0]
            try:
                if msg not in server_command:
                    Logging.error(f"Unknown command {msg}")
                    clientsocket.send("failed".encode("utf-8"))
                else:
                    response = server_command[msg]["command"](*argument)
                    if isinstance(response, str):
                        response = response.encode("utf-8")
                    clientsocket.send(response)
            except Exception as e:
                # show backtrace
                Logging.error("Server failed", e)
                Logging.error("Config", Config.get())

                clientsocket.send("failed".encode("utf-8"))
            finally:
                clientsocket.close()

    threading.Thread(target=server_loop).start()


def send_msg_to_server(msg):
    # create a socket object
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # get local machine name
    host = socket.gethostname()
    port = Config.get()["server"]["port"]

    try:
        # connection to hostname on the port.
        client.connect((host, port))
    except Exception as e:
        Logging.error("Client failed to connect to server", e)
        exit(1)

    # Receive no more than 1024 bytes
    client.send(msg.encode("utf-8"))
    response = client.recv(1024).decode("utf-8")
    client.close()

    return response



import time

from monitor.log import Logging


def is_in_time_range(time_range_str):
    def parse_time_range(time_range):
        pairs = time_range.split(",")
        result = []
        for pair in pairs:
            start, end = pair.split("-")
            result.append((start, end))
        return result

    time_range = parse_time_range(time_range_str)
    Logging.debug(f"Time range: {time_range}")

    current_time = time.localtime()
    current_time_str = time.strftime("%H:%M", current_time)
    Logging.debug(f"Current time: {current_time_str}")
    for time_pair in time_range:
        Logging.debug(f"Range pair: {time_pair}")
        start, end = time.strftime(
            "%H:%M", time.strptime(time_pair[0], "%H:%M")
        ), time.strftime("%H:%M", time.strptime(time_pair[1], "%H:%M"))
        if start <= current_time_str <= end:
            Logging.debug("In range")
            return True
    Logging.debug("Not in range")
    return False


class PolicyParams(object):
    frames_count = 0


def easy_policy(config, capture, video):
    if PolicyParams.frames_count == 0:
        PolicyParams.frames_count = len(video.all_frames())
        Logging.debug(f"Initial frames count: {PolicyParams.frames_count}")

    easy_config = config["easy_policy"]
    frames_per_video = easy_config["frames_per_video"]
    frames_interval = easy_config["frames_interval"]

    Logging.debug(f"Frames count: {PolicyParams.frames_count}, frames per video: {frames_per_video}, frames interval: {frames_interval}")
    if PolicyParams.frames_count >= frames_per_video and is_in_time_range(
        config["generate_time"]
    ):
        Logging.debug("Generate video")
        video.generate_video()
        PolicyParams.frames_count = 0

    # update the actual frames count, but not frequently
    if PolicyParams.frames_count == (frames_per_video // 2):
        Logging.debug(f"Before update frames count: {PolicyParams.frames_count}")
        PolicyParams.frames_count = len(video.all_frames())
        Logging.debug(f"After update frames count: {PolicyParams.frames_count}")

    capture.capture()
    PolicyParams.frames_count += 1
    time.sleep(frames_interval)



import cv2
import os
import datetime
import time
import pathlib

from monitor.log import Logging


class Capture(object):
    captured_frames_ = 0
    max_retry_ = 10

    @staticmethod
    def captured_frames():
        return Capture.captured_frames_

    def __init__(self, video_path_for_debug="", config=None):
        self.config_ = config
        self.video_path_for_debug_ = video_path_for_debug
        self.cap_ = None
        self.live_ = True
        if pathlib.Path(self.video_path_for_debug_).exists():
            Logging.info(f"Load frames from {self.video_path_for_debug_}")
            self.cap_ = cv2.VideoCapture(self.video_path_for_debug_)
            self.live_ = False

    def update_config(self, config):
        self.config_ = config

    def capture(self):
        max_retry = Capture.max_retry_

        while True:
            # no file specified, use camera
            if self.cap_ is None:
                self.cap_ = cv2.VideoCapture(self.config_["camera_id"], cv2.CAP_V4L2)
            if not self.cap_.isOpened():
                Logging.error("Capture not valid")
                raise Exception("Capture not valid")

            ret, frame = self.cap_.read()
            if not ret:
                time.sleep(1)
                max_retry -= 1
                if max_retry == 0:
                    Logging.error(f"Open camera failed, retrying {max_retry}")
                    raise Exception("Open camera failed")
                else:
                    Logging.error(f"Open camera failed, retrying {max_retry}")
                continue

            capture_date = datetime.datetime.now()
            # 2021/01/01 00:00:00.000
            frame_date = capture_date.strftime("%Y/%m/%d %H:%M:%S.%f")
            frame_date = frame_date[:-3]
            Logging.info(f"Captured {frame_date}")
            cv2.putText(
                frame,
                frame_date,
                (10, 50),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                # yellow
                (0, 255, 255),
                2,
                cv2.LINE_AA,
            )

            # set jpeg quality
            capture_date = capture_date.strftime("%Y%m%d%H%M%Ss%f")
            capture_date = capture_date[:-3]
            cv2.imwrite(
                f"{self.config_['frames_dir']}/{capture_date}.jpeg",
                frame,
                [int(cv2.IMWRITE_JPEG_QUALITY), self.config_["quality"]],
            )
            Capture.captured_frames_ += 1

            if self.live_:
                self.cap_.release()
            break



import sys
import pathlib
import datetime
import os


class Logging(object):
    log_levels_ = ["FATAL", "ERROR", "WARN", "INFO", "DEBUG"]
    latest_log_file_ = None
    log_level_ = 3
    config_ = None

    @staticmethod
    def update_config(config):
        Logging.config_ = config

    @staticmethod
    def update_log_level(log_level):
        Logging.log_level_ = Logging.log_levels_.index(log_level)

    @staticmethod
    def info(*args, **kwargs):
        Logging.log_("INFO", *args, **kwargs)

    @staticmethod
    def error(*args, **kwargs):
        Logging.log_("ERROR", *args, **kwargs)

    @staticmethod
    def debug(*args, **kwargs):
        Logging.log_("DEBUG", *args, **kwargs)

    @staticmethod
    def warn(*args, **kwargs):
        Logging.log_("WARN", *args, **kwargs)

    @staticmethod
    def fatal(*args, **kwargs):
        Logging.log_("FATAL", *args, **kwargs)

    @staticmethod
    def write(tag, *args, **kwargs):
        if Logging.should_generate_new_log_file_():
            Logging.get_latest_log_file_()

        # caller_name = sys._getframe(2).f_code.co_name
        caller_file = sys._getframe(2).f_code.co_filename
        caller_line = sys._getframe(2).f_lineno
        current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        caller_file = pathlib.Path(caller_file).name
        caller_info = f"{caller_file}:{caller_line}"
        prefix_info = f"{current_time} [{os.getpid()}] {tag} {caller_info}"
        message = f"{prefix_info} {' '.join([str(arg) for arg in args])} {' '.join([f'{key}={value}' for key, value in kwargs.items()])}"
        with open(Logging.latest_log_file_, "a") as f:
            print(message, file=f)
        return message

    @staticmethod
    def get_latest_log_file_():
        def generate_log_file_by_date():
            current_date = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            log_path = pathlib.Path(Logging.config_["log_dir"]) / f"{current_date}.log"
            log_path.touch()
            return log_path

        log_files = sorted(pathlib.Path(Logging.config_["log_dir"]).glob("*.log"))
        if not log_files:
            Logging.latest_log_file_ = generate_log_file_by_date()
            return
        Logging.latest_log_file_ = log_files[-1]
        # size > 1 MB
        if Logging.latest_log_file_.stat().st_size > 1e6:
            Logging.latest_log_file_ = generate_log_file_by_date()
        # max 10 files
        if len(log_files) > 10:
            log_files[0].unlink()

    @staticmethod
    def should_generate_new_log_file_():
        if not Logging.latest_log_file_:
            Logging.get_latest_log_file_()
        if Logging.latest_log_file_.stat().st_size > 1e6:
            return True
        return False

    @staticmethod
    def log_(level, *args, **kwargs):
        Logging.check_config_ready_()
        if Logging.log_levels_.index(level) > Logging.log_level_:
            return
        print(Logging.write(f"[{level}]", *args, **kwargs))

    @staticmethod
    def check_config_ready_():
        if not Logging.config_:
            raise Exception("Logging config not ready")



from monitor.log import Logging
from monitor.config import Config
from monitor.capture import Capture
from monitor.video import Video
from monitor.policy import *


def client_send_msg_to_server(msg, *args):
    from monitor.server import send_msg_to_server

    send_msg = " ".join([msg, *args])
    response = send_msg_to_server(send_msg)

    Logging.write("[INFO ]", "Client received", response)
    if response:
        print(response)


class Command(object):
    command_ = {}

    @staticmethod
    def get():
        return Command.command_

    @staticmethod
    def add_command(name, help="", server=False):
        if not Command.command_:
            from monitor.server import ServerCommand

            server_command = ServerCommand.get()

            for srever_command_name, value in server_command.items():
                Command.command_[srever_command_name] = {
                    "command": client_send_msg_to_server,
                    "help": value["help"],
                    "server": True,
                }
                Logging.debug(
                    f"Command[{srever_command_name}] {Command.command_[srever_command_name]}"
                )

        def decorator(func):
            Command.command_[name] = {
                "command": func,
                "help": help,
                "server": server,
            }
            Logging.debug(f"Command[{name}] {Command.command_[name]}")
            return func

        return decorator


@Command.add_command("help", "Print help")
def help(command_name):
    Logging.debug(f"Command[{command_name}]")
    help_str = """Usage: python3 -m monitor <command> [arguments]
Commands:
"""
    for name, value in Command.get().items():
        help_str += f"    {name}: {value['help']}\n"
    print(help_str)


@Command.add_command("development", "Development mode, show more logs")
def development(command_name, *args):
    Logging.update_log_level("DEBUG")
    Logging.debug(f"Command[{command_name}]")
    server("server", *args)


@Command.add_command("server", "Start server [video_path], default video_path is empty")
def server(command_name, *args):
    Logging.debug(f"Command[{command_name}]")
    Logging.info("Starting")
    config = Config.get()
    policy = config["policy"]
    Logging.info(f"Using policy {policy}")
    # str to function
    policy = globals()[policy]

    video_path_for_debug = "" if len(args) == 0 else args[0]
    video = Video(config)
    capture = Capture(config=config, video_path_for_debug=video_path_for_debug)

    from monitor.server import start_server, should_stop

    start_server()
    while not should_stop():
        try:
            config = Config.get()
            video.update_config(config)
            capture.update_config(config)

            policy(config, capture, video)
        except Exception as e:
            Logging.error(e)
            # backtrace
            import traceback

            traceback.print_exc()
            client_send_msg_to_server("stop")
    Logging.info("Stopped")
    exit(0)



import cv2
import time
import pathlib

from monitor.log import Logging


class Video(object):
    # when generating video, capture will still run
    video_in_progress_ = False
    generated_videos_ = []

    @staticmethod
    def generated_videos():
        return Video.generated_videos_

    def __init__(self, config=None):
        self.config_ = config

    def update_config(self, config):
        self.config_ = config

    def all_frames(self) -> list:
        frames_dir = self.config_["frames_dir"]
        # when generating video, capture will still run, so do not check frames dir
        if Video.video_in_progress_:
            return []

        target_frames_suffix = {".jpeg", ".jpg", ".png"}
        frames = []
        for frame_path in pathlib.Path(frames_dir).glob("*"):
            if frame_path.suffix in target_frames_suffix:
                frames.append(frame_path)
        return frames

    def generate_video(self):
        # only one video can be generated at a time
        if Video.video_in_progress_:
            Logging.debug("Video in progress, skip generating video")
            return

        Logging.info("Generating video")
        frames, frames_date_range = self.load_frames_()
        Logging.info(f"Total frames: {len(frames)}, date range: {frames_date_range}")
        if not frames:
            return

        Video.video_in_progress_ = True
        import threading

        threading.Thread(
            target=self.generate_video_from_frames_,
            args=(
                frames,
                frames_date_range,
                self.config_["video_dir"],
                self.config_["fps"],
            ),
        ).start()

    def load_frames_(self):
        frames_dir = self.config_["frames_dir"]
        frames = []
        first_frame_date = None
        last_frame_date = None
        sorted_frame_paths = sorted(self.all_frames())
        for frame_path in sorted_frame_paths:
            frame = str(frame_path)
            frames.append(frame)
            if not first_frame_date:
                first_frame_date = frame_path.stem
            last_frame_date = frame_path.stem
        frames_date_range = f"{first_frame_date}-{last_frame_date}"
        return frames, frames_date_range

    def generate_video_from_frames_(self, frames, frames_date_range, video_dir, fps):
        base_frame = cv2.imread(frames[0])
        height, width, _ = base_frame.shape
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        output_path = pathlib.Path(video_dir) / f"{frames_date_range}.mp4"
        if output_path.exists():
            Logging.info(f"Removing existing video {output_path}")
            output_path.unlink()
        Logging.info(f"Generating video {output_path}")
        video = cv2.VideoWriter(str(output_path), fourcc, fps, (width, height))

        skip_write = False
        frame_save = self.config_["frames_save"]
        for frame_path in frames:
            frame = cv2.imread(frame_path)
            if not frame_save:
                pathlib.Path(frame_path).unlink()
            if not skip_write:
                video.write(frame)
        video.release()
        Logging.info(f"Video generated {output_path}")

        Video.generated_videos_.append(frames_date_range)
        self.remove_outdated_videos_(video_dir)

        Video.video_in_progress_ = False

    def remove_outdated_videos_(self, video_dir):
        video_paths = sorted(pathlib.Path(video_dir).glob("*.mp4"))
        filtered_video_paths = []

        import re

        # the pattern is like "20240107153612s239-20240107153637s319.mp4"
        for video_path in video_paths:
            match = re.search(r"(\d+s\d+)-(\d+s\d+)", video_path.stem)
            if not match:
                continue
            start, end = video_path.stem.split("-")
            if len(start) != 18 or len(end) != 18:
                continue
            filtered_video_paths.append(video_path)

        Logging.debug(f"Found {len(filtered_video_paths)} videos")

        sorted_video_paths = sorted(filtered_video_paths)
        for video_path in sorted_video_paths[: -self.config_["iterations"]]:
            Logging.info(f"Removing outdated video {video_path}")
            video_path.unlink()



import json
import os
import pathlib


class Config(object):
    raw_config_ = {
        "camera_id": 0,
        "video_dir": "$HOME/Videos",
        "frames_dir": "$HOME/Pictures/work-monitor",
        "config_dir": "$HOME/.work-monitor",
        "log_dir": "$HOME/.work-monitor/log",
        "fps": 60,
        "quality": 75,
        "frames_save": False,
        "generate_time": "23:00-00:00,00:00-06:00",
        "iterations": 10,
        "policy": "easy_policy",
        "easy_policy": {"frames_interval": 10, "frames_per_video": 1000},
        "server": {
            "port": 22311,
        },
    }
    config_ = None

    @staticmethod
    def get():
        return Config.config_.copy()

    @staticmethod
    def set(value):
        Config.config_ = value

    @staticmethod
    def init() -> None:
        if Config.config_ is not None:
            return

        Config.config_ = Config.raw_config_.copy()
        try:
            Config.config_ = Config.initialize_config(Config.config_)
            Config.verify_config(Config.config_)
        except Exception as e:
            print(e)
            choise = input("Reset config? (y/n)")
            if choise == "y":
                print("Resetting config")
                Config.config_ = Config.update_better_config(
                    Config.config_, Config.raw_config_
                )
                print("Raw config: ")
                print(json.dumps(Config.raw_config_, indent=4))
                print("Current config: ")
                print(json.dumps(Config.config_, indent=4))
                Config.config_ = Config.initialize_config(Config.config_, force=True)
            else:
                print("Using raw config")
                Config.config_ = Config.raw_config_
        Config.config_ = Config.preprocess_config(Config.config_)
        Config.verify_config(Config.config_)

    @staticmethod
    def update_config(config, other_config):
        config.update(other_config)
        return config

    @staticmethod
    def parse_time_range(config, key):
        pairs = config[key].split(",")
        result = []
        for pair in pairs:
            start, end = pair.split("-")
            start = list(map(int, start.split(":")))
            end = list(map(int, end.split(":")))
            result.append((start, end))
        return result

    @staticmethod
    def preprocess_config(config):
        def replace_env_var(config):
            for key, value in config.items():
                if isinstance(value, str):
                    config[key] = value.replace("$HOME", os.environ["HOME"])
                elif isinstance(value, dict):
                    replace_env_var(value)

        def replace_stringnumber_to_number(config):
            for key, value in config.items():
                if isinstance(value, str):
                    if value.isdigit():
                        config[key] = int(value)
                    elif value.replace(".", "", 1).isdigit():
                        config[key] = float(value)
                elif isinstance(value, dict):
                    replace_stringnumber_to_number(value)

        def create_dir_if_not_exist(config):
            for key, value in config.items():
                if isinstance(value, str) and key.endswith("_dir"):
                    config[key] = pathlib.Path(value)
                    pathlib.Path(value).mkdir(parents=True, exist_ok=True)
                elif isinstance(value, dict):
                    create_dir_if_not_exist(value)

        def reset_value_range(config, key, min_value, max_value):
            if key not in config:
                return
            if config[key] < min_value:
                config[key] = min_value
            elif config[key] > max_value:
                config[key] = max_value

        replace_env_var(config)
        replace_stringnumber_to_number(config)
        create_dir_if_not_exist(config)
        reset_value_range(config, "quality", 0, 100)
        return config

    @staticmethod
    def stringify_config(config):
        def stringify(config):
            for key, value in config.items():
                if isinstance(value, pathlib.Path):
                    config[key] = str(value)
                elif isinstance(value, dict):
                    stringify(value)

        stringify(config)
        return config

    @staticmethod
    def initialize_config(config, force=False):
        Config.preprocess_config(config)
        config_path = config["config_dir"] / "config.json"
        if not config_path.exists() or force:
            config_path.parent.mkdir(parents=True, exist_ok=True)
            json.dump(Config.stringify_config(config), open(config_path, "w"), indent=4)
        config = json.load(open(config_path))
        return config

    @staticmethod
    def verify_config(config):
        raw_config_keys = set()
        config_keys = set()

        def append_sub_keys(keys, config, prefix=""):
            for key, value in config.items():
                keys.add(f"{prefix}{key}")
                if isinstance(value, dict):
                    append_sub_keys(keys, value, prefix=f"{prefix}{key}_")

        append_sub_keys(raw_config_keys, Config.raw_config_)
        append_sub_keys(config_keys, config)

        if raw_config_keys != config_keys:
            not_match = raw_config_keys ^ config_keys
            raise Exception("Config keys not match", not_match)

    @staticmethod
    def update_better_config(config, raw_config):
        def update_better_config_helper(config, raw_config):
            for key, value in config.items():
                if key not in raw_config:
                    config.pop(key)
                    continue
                if type(value) != type(raw_config[key]):
                    config[key] = raw_config[key]
                if isinstance(value, dict):
                    update_better_config_helper(value, raw_config[key])

        update_better_config_helper(config, raw_config)

        def add_missing_keys(config, raw_config):
            for key, value in raw_config.items():
                if key not in config:
                    config[key] = value
                if isinstance(value, dict):
                    add_missing_keys(config[key], value)

        add_missing_keys(config, raw_config)

        return config



import sys
from monitor.config import Config
from monitor.log import Logging


def main():
    Config.init()
    Logging.update_config(Config.get())

    command_section = sys.argv[1] if len(sys.argv) > 1 else "help"
    command_argument = sys.argv[2:] if len(sys.argv) > 2 else []

    from monitor.command import Command

    command = Command.get()
    if command_section not in command:
        raise ValueError(
            f"Section {command_section} not found, valid keys are {command.keys()}"
        )
    command[command_section]["command"](command_section, *command_argument)


if __name__ == "__main__":
    main()
